﻿// ItemsControlTransitions by Charles Petzold, March 2010
using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;

namespace ItemsControlTransitions
{
    public partial class MainPage : UserControl
    {
        ObservableCollection<ProduceItem> marketCollection = new ObservableCollection<ProduceItem>();
        ObservableCollection<ProduceItem> basketCollection = new ObservableCollection<ProduceItem>();
        ItemContainerGenerator marketGenerator;
        ItemContainerGenerator basketGenerator;

        bool isDragging;
        bool isAnimating;
        FrameworkElement marketItemContainer;
        FrameworkElement basketItemContainer;
        Point ptDragCtrlStart;
        Point ptMousePrevious;
        object itemBeingTransferred;

        public static readonly DependencyProperty ScrollProperty =
            DependencyProperty.Register("Scroll",
                typeof(double),
                typeof(MainPage),
                new PropertyMetadata(OnScrollChanged));

        public MainPage()
        {
            InitializeComponent();

            // Associate collections with ItemsControls
            market.ItemsSource = marketCollection;
            basket.ItemsSource = basketCollection;

            // Obtain ItemContainerGenerators
            marketGenerator = market.ItemContainerGenerator;
            basketGenerator = basket.ItemContainerGenerator;

            // Get all the data
            ProduceItemsPresenter presenter = new ProduceItemsPresenter();
            presenter.PropertyChanged += delegate(object sender, PropertyChangedEventArgs args)
            {
                if (args.PropertyName == "ProduceItems")
                {
                    // Fill marketCollection with all the data
                    foreach (ProduceItem item in presenter.ProduceItems.Items)
                        marketCollection.Add(item);
                }
            };
        }

        public double Scroll
        {
            set { SetValue(ScrollProperty, value); }
            get { return (double)GetValue(ScrollProperty); }
        }

        void OnMarketItemsControlMouseLeftButtonDown(object sender, MouseButtonEventArgs args)
        {
            if (isAnimating)
                return;

            // Get the container of the item that was clicked
            DependencyObject container;
            object item = GetClickedItem(args.OriginalSource as DependencyObject, marketGenerator,
                                         typeof(ProduceItem), out container);

            // The container should be a ContentPresenter
            if (container is FrameworkElement)
            {
                // Hide the container but leave the item in the ItemsControl
                marketItemContainer = container as FrameworkElement;
                marketItemContainer.Opacity = 0;

                // Set the ContentControl content and position to the item
                dragControl.Content = item;
                ptDragCtrlStart = marketItemContainer.TransformToVisual(this).Transform(new Point());

                // Enable the ContentControl to drag across screen
                Canvas.SetLeft(dragControl, ptDragCtrlStart.X);
                Canvas.SetTop(dragControl, ptDragCtrlStart.Y);
                dragControl.Visibility = Visibility.Visible;

                // Let's go!
                ptMousePrevious = args.GetPosition(this);
                CaptureMouse();
                isDragging = true;
                args.Handled = true;
            }
        }

        protected override void OnMouseMove(MouseEventArgs args)
        {
            if (isDragging)
            {
                // Move the ContentControl
                Point ptMouse = args.GetPosition(this);
                Canvas.SetLeft(dragControl, Canvas.GetLeft(dragControl) + ptMouse.X - ptMousePrevious.X);
                Canvas.SetTop(dragControl, Canvas.GetTop(dragControl) + ptMouse.Y - ptMousePrevious.Y);
                ptMousePrevious = ptMouse;
            }

            base.OnMouseMove(args);
        }

        protected override void OnMouseLeftButtonUp(MouseButtonEventArgs args)
        {
            if (isDragging)
            {
                // Find intersection of basket ItemsControl and dragControl
                Point ptBasket = basketScrollViewer.TransformToVisual(this).Transform(new Point());
                Rect rectBasket = new Rect(ptBasket, new Size(basket.ActualWidth, basketScrollViewer.ActualHeight));
                Point ptDragCtrl = dragControl.TransformToVisual(this).Transform(new Point());                      // or Canvas.Left and Canvas.Top
                Rect rectDragCtrl = new Rect(ptDragCtrl, new Size(dragControl.ActualWidth, dragControl.ActualHeight));
                rectBasket.Intersect(rectDragCtrl);

                // If no intersection, return the item to its origin
                if (rectBasket.IsEmpty)
                {
                    // Initialize the animations
                    returnToOriginAnimaX.To = ptDragCtrlStart.X;
                    returnToOriginAnimaY.To = ptDragCtrlStart.Y;

                    // Begin the storyboard
                    isAnimating = true;
                    returnToOriginStoryboard.Begin();
                }
                else
                {
                    // Get the item from the dragControl and save in a field
                    itemBeingTransferred = dragControl.Content;

                    // Add the item to the basket
                    Point ptDestination;
                    int index = GetBasketDestinationIndex(rectDragCtrl, out ptDestination);
                    basketCollection.Insert(index, itemBeingTransferred as ProduceItem);

                    // Make the item's container invisible 
                    basketItemContainer = basketGenerator.ContainerFromItem(itemBeingTransferred) as FrameworkElement;
                    basketItemContainer.Opacity = 0;

                    // Calculate ScrollViewer scrolling animation
                    double deficiencyOnTop = Math.Min(0, ptDestination.Y - ptBasket.Y);

                    double excessBeyondBottom = Math.Max(0, 
                                    ptDestination.Y + dragControl.ActualHeight - 
                                        (ptBasket.Y + basketScrollViewer.ActualHeight));

                    double offset = deficiencyOnTop != 0 ? deficiencyOnTop : excessBeyondBottom;

                    ptDestination.Y -= offset;
                    scrollItemsControlAnima.From = basketScrollViewer.VerticalOffset;
                    scrollItemsControlAnima.To = basketScrollViewer.VerticalOffset + offset;

                    // Continue initializing the animations
                    Storyboard.SetTarget(closeUpMarketHeightAnima, marketItemContainer);
                    Storyboard.SetTarget(openUpBasketHeightAnima, basketItemContainer);
                    moveToBasketAnimaX.To = ptDestination.X;
                    moveToBasketAnimaY.To = ptDestination.Y;

                    // Begin them!
                    isAnimating = true;
                    transferToBasketStoryboard.Begin();
                }

                isDragging = false;
                ReleaseMouseCapture();
            }
            base.OnMouseLeftButtonUp(args);
        }

        // Scroll dependency property changed handler
        static void OnScrollChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as MainPage).OnScrollChanged(args);
        }

        void OnScrollChanged(DependencyPropertyChangedEventArgs args)
        {
            if (isAnimating)
                basketScrollViewer.ScrollToVerticalOffset((double)args.NewValue);
        }

        // Two Storyboard Completed event handlers
        void OnReturnToOriginStoryboardCompleted(object sender, EventArgs args)
        {
            // Restore the opacity of item in market
            marketItemContainer.Opacity = 1;

            // Hide the ContentControl
            dragControl.Visibility = Visibility.Collapsed;

            // Stop the storyboard to free up animations
            isAnimating = false;
            (sender as Storyboard).Stop();
        }

        void OnTransferToBasketStoryboardCompleted(object sender, EventArgs args)
        {
            // Remove item from market
            marketCollection.Remove(itemBeingTransferred as ProduceItem);

            // Set opacity of item in basket
            basketItemContainer.Opacity = 1;

            // Hide the ContentControl
            dragControl.Visibility = Visibility.Collapsed;

            // Stop the storyboard to free up animations
            isAnimating = false;
            (sender as Storyboard).Stop();
        }

        object GetClickedItem(DependencyObject clickSource, ItemContainerGenerator generator,
                      Type itemType, out DependencyObject container)
        {
            if (clickSource == null)
            {
                container = null;
                return null;
            }

            object item = generator.ItemFromContainer(clickSource);

            if (item != null && item.GetType() == itemType)
            {
                container = clickSource;
                return item;
            }

            return GetClickedItem(VisualTreeHelper.GetParent(clickSource), generator, itemType, out container);
        }

        int GetBasketDestinationIndex(Rect rectDragCtrl, out Point ptDestination)
        {
            Point ptLowerLeft = basket.TransformToVisual(this).Transform(new Point());

            for (int index = 0; index < basketCollection.Count; index++)
            {
                FrameworkElement basketItemContainer = basketGenerator.ContainerFromIndex(index) as FrameworkElement;
                Point ptUpperLeft = basketItemContainer.TransformToVisual(this).Transform(new Point());
                Point ptUpperRight = basketItemContainer.TransformToVisual(this).Transform(new Point(basketItemContainer.ActualWidth, 0));
                ptLowerLeft = basketItemContainer.TransformToVisual(this).Transform(new Point(0, basketItemContainer.ActualHeight));

                if (rectDragCtrl.Contains(ptUpperLeft) || rectDragCtrl.Contains(ptUpperRight))
                {
                    ptDestination = ptUpperLeft;
                    return index;
                }
            }

            ptDestination = ptLowerLeft;
            return basketCollection.Count;
        }
    }
}
